"""
Provides middleware to support classic CUL/Apache log format.

For accurate request metrics, apply this middleware before any others.
"""

from typing import Type, Callable, List, Iterable, Tuple
from datetime import datetime
from pytz import timezone
from .base import BaseMiddleware

EASTERN = timezone('US/Eastern')

# The uwsgi package is injected into the Python environment by uWSGI, and is
# not available otherwise. If we can't import the package, there's not much
# else to do.
try:
    import uwsgi

    class ClassicLogsMiddleware(BaseMiddleware):
        """Add params to uwsgi to support classic CUL/Apache log format."""

        def before(self, environ: dict, start_response: Callable) \
                -> Tuple[dict, Callable]:
            """
            Inject request time in CUL classic format, and start TTFB clock.

            Parameters
            ----------
            environ : dict
                WSGI request environment.
            start_response : function
                Function used to begin the HTTP response. See
                https://www.python.org/dev/peps/pep-0333/#the-start-response-callable

            Returns
            -------
            dict
                Request environment (``environ``).
            function
                ``start_response`` function.
            """
            self.start = datetime.now()

            # Express the time that the request was received in the classic
            # format.
            rtime = datetime.now(tz=EASTERN).strftime('%d/%b/%Y:%H:%M:%S %z')
            uwsgi.set_logvar('rtime', rtime)
            return environ, start_response

        def after(self, response: Iterable) -> Iterable:
            """
            Set TTFB log variable.

            Parameters
            ----------
            response : iterable
                Iterable that generates the HTTP response. See
                https://www.python.org/dev/peps/pep-0333/#the-application-framework-side

            Returns
            -------
            iterable
                Iterable that generates the HTTP response. See
                https://www.python.org/dev/peps/pep-0333/#the-application-framework-side
            """
            # This is a close approximation of "TTFB" (time from the
            # request received to the start of the response).
            ttfb = (datetime.now() - self.start).microseconds
            uwsgi.set_logvar('ttfb', str(ttfb).encode('ascii'))
            return response

except ImportError as e:     # Not running in uWSGI.
    # BaseMiddleware does nothing.
    ClassicLogsMiddleware = BaseMiddleware  # type: ignore



"""
Base arXiv package.

Each component of this project **must** meet all of the following criteria:

1. It is likely that the component will be utilized in many or all arXiv
   services.
2. Once stable, it is unlikely to change often.
3. It is unlikely that implementing new features in specific services
   would require changes to the component.
4. When a component does change, it **must** change in the same way for all of
   the services that use it.

"""



"""Utility functions for arxiv.search."""
# pylint: disable=C0330

import re

__all__ = ('parse_arxiv_id', )

CATEGORIES = [
    "acc-phys", "adap-org", "alg-geom", "ao-sci", "astro-ph", "atom-ph",
    "bayes-an", "chao-dyn", "chem-ph", "cmp-lg", "comp-gas", "cond-mat", "cs",
    "dg-ga", "funct-an", "gr-qc", "hep-ex", "hep-lat", "hep-ph", "hep-th",
    "math", "math-ph", "mtrl-th", "nlin", "nucl-ex", "nucl-th", "patt-sol",
    "physics", "plasm-ph", "q-alg", "q-bio", "quant-ph", "solv-int",
    "supr-con", "eess", "econ"
]

ARXIV_REGEX = (
    "^(ar[xX]iv:)?((?:(?:(?:%s)(?:[.][A-Z]{2})?/[0-9]{2}(?:0[1-9]|1[0-2])"
    "\\d{3}(?:[vV]\\d+)?))|(?:(?:[0-9]{2}(?:0[1-9]|1[0-2])[.]"
    "\\d{4,5}(?:[vV]\\d+)?)))$" % '|'.join(CATEGORIES)
)


def parse_arxiv_id(value: str) -> str:
    """
    Parse arxiv id from string.

    Raises `ValidationError` if no arXiv ID.
    """
    m = re.search(ARXIV_REGEX, value)
    if not m:
        raise ValueError('Not a valid arXiv ID')
    return m.group(2)



"""
arXiv base Flask components.

Provides :class:`.Base`, which attaches base templates, static assets, global
context processors, and exception handlers to a :class:`flask.Flask` app
instance.

Intended for use in an application factory. For example:

.. code-block:: python

   python
   from flask import Flask
   from arxiv.base import Base
   from someapp import routes


   def create_web_app() -> Flask:
      app = Flask('someapp')
      app.config.from_pyfile('config.py')

      Base(app)   # Registers the base/UI blueprint.
      app.register_blueprint(routes.blueprint)    # Your blueprint.
   return app


"""

from typing import Optional
from flask import Blueprint, Flask

from arxiv.base.context_processors import config_url_builder
from arxiv.base import exceptions
from arxiv.base.converter import ArXivConverter


class Base(object):
    """Attaches a base UI blueprint and context processors to an app."""

    def __init__(self, app: Optional[Flask] = None) -> None:
        """Initialize ``app`` with base blueprint."""
        if app is not None:
            self.init_app(app)

    def init_app(self, app: Flask) -> None:
        """Create and register the base UI blueprint."""
        # Attach the arXiv identifier converter for URLs with IDs.
        app.url_map.converters['arxiv'] = ArXivConverter

        # The base blueprint attaches static assets and templates.
        blueprint = Blueprint(
            'base',
            __name__,
            template_folder='templates',
            static_folder='static',
            static_url_path=app.static_url_path + '/base'
        )
        app.register_blueprint(blueprint)

        # Register base context processors (e.g. to inject global URLs).
        app.context_processor(config_url_builder)

        # Register base exception handlers.
        for error, handler in exceptions.get_handlers():
            app.errorhandler(error)(handler)



"""Install arXiv-base as an importable package."""

from setuptools import setup, find_packages


setup(
    name='arxiv-base',
    version='0.5.1',
    packages=find_packages(exclude=['tests.*']),
    zip_safe=False,
    include_package_data=True
)



"""Tests for :mod:`arxiv.base.context_processors`."""

from unittest import TestCase, mock
from flask import Flask

from arxiv.base.context_processors import config_url_builder
from arxiv.base.exceptions import ConfigurationError


class TestConfigURLBuilderContextProcessor(TestCase):
    """config_url_builder() injects a URL factory into the response context."""

    def test_config_url_builder_returns_a_function(self):
        """config_url_builder() should return a dict with config_url()."""
        extra_context = config_url_builder()
        self.assertIsInstance(extra_context, dict, "Should return a dict")
        self.assertIn('config_url', extra_context,
                      "Should contain `config_url` key")
        self.assertTrue(hasattr(extra_context['config_url'], '__call__'),
                        "Value for ``config_url`` should be callable.")

    def test_config_url_key_exists(self):
        """config_url() is called for an URL that is configured on the app."""
        app = Flask('foo')
        app.config['ARXIV_FOO_URL'] = 'https://foo.arxiv.org'
        with app.app_context():
            extra_context = config_url_builder()
            config_url = extra_context['config_url']
            url = config_url('foo')

        self.assertEqual(url, app.config['ARXIV_FOO_URL'],
                         "Should return configured URL")

    @mock.patch('arxiv.base.context_processors.config')
    def test_config_url_key_not_on_current_app(self, mock_base_config):
        """config_url() is called for an URL that is not on the current app."""
        mock_base_config.ARXIV_FOO_URL = 'https://bar.arxiv.org'
        app = Flask('foo')
        with app.app_context():
            extra_context = config_url_builder()
            config_url = extra_context['config_url']
            url = config_url('foo')

        self.assertEqual(url, 'https://bar.arxiv.org',
                         "Should return URL configured in arxiv base ")

    def test_config_url_key_not_set_anywhere(self):
        """config_url() is called for an URL that is not configured."""
        app = Flask('foo')
        with app.app_context():
            extra_context = config_url_builder()
            config_url = extra_context['config_url']
            with self.assertRaises(ConfigurationError):
                config_url('foo')



"""
Provides routes for verifying base templates and exception handling.

The blueprint instantiated here is **not** for use in a production application;
it is not attached by :class:`arxiv.base.Base`.
"""

from typing import Any, Tuple, Callable, Dict
from flask import Blueprint, render_template, current_app, make_response, \
    Response

from arxiv import status
from arxiv.base.exceptions import NotFound, Forbidden, Unauthorized, \
    MethodNotAllowed, RequestEntityTooLarge, BadRequest, InternalServerError


blueprint = Blueprint('ui', __name__, url_prefix='')


@blueprint.route('/test', methods=['GET'])
def test_page() -> Response:
    """Render the test page."""
    rendered = render_template("base/base.html", pagetitle='Home')
    response = make_response(rendered, status.HTTP_200_OK)
    return response


@blueprint.route('/404', methods=['GET'])
def test_404() -> Response:
    """Test the 404 error page."""
    raise NotFound()


@blueprint.route('/403', methods=['GET'])
def test_403() -> Response:
    """Test the 403 error page."""
    raise Forbidden()


@blueprint.route('/401', methods=['GET'])
def test_401() -> Response:
    """Test the 401 error page."""
    raise Unauthorized()


@blueprint.route('/405', methods=['GET'])
def test_405() -> Response:
    """Test the 405 error page."""
    raise MethodNotAllowed()


@blueprint.route('/413', methods=['GET'])
def test_413() -> Response:
    """Test the 413 error page."""
    raise RequestEntityTooLarge()


@blueprint.route('/400', methods=['GET'])
def test_400() -> Response:
    """Test the 400 error page."""
    raise BadRequest()


@blueprint.route('/500', methods=['GET'])
def test_500() -> Response:
    """Test the 500 error page."""
    raise InternalServerError()
